#ifdef USE_AUDIO
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#ifdef USE_SDL1
#include <SDL/SDL.h>
#endif
#ifdef USE_SDL2
#include <SDL2/SDL.h>
#endif

#ifdef USE_NLLIBC
#include <nllibc.h>
#endif

#include "config.h"
#include "const.h"
#include "nlllib.h"
#include "memory.h"
#include "audio.h"
#include "sound.h"
#include "sdl.h"

struct osc {
  struct osc *next;
  int index;
  int type;
  int amp;
  int freq;
  int phase;
  unsigned int flags;

  int outnum;
  struct {
    int amp;
    struct osc *osc;
  } outs[AUDIO_OSC_OUTMAX];

  struct {
    int num;
    int amp;
    struct osc *osc;
  } in;

  struct {
    int amp;
  } out;

  struct {
    int num;
    struct envpoint {
      int step;
      int amp;
    } points[AUDIO_OSC_ENVMAX];
  } envelope;
};

static struct oscset {
  int num;
  struct osc *oscs;
} oscsets[AUDIO_OSCSET_MAXNUM];

struct oscregs {
  struct oscregs *next;

  struct osc *osc;

  struct {
    int amp;
  } out;

  struct {
    int index;
  } envelope;
};

struct samplestep {
  int sec;
  int count;
};

struct player {
  struct player *next;
  struct player *nextplayer;
  struct oscregs *pool;
  int duplicate;

  struct player *head;
  int oscset_index;
  struct oscregs *oscregs;
  int amp;
  int freq;
  struct samplestep step;
  int stepbase;
  int steplen;
};

static struct player *playeres = NULL;

static struct osc *osc_pool = NULL;
static struct oscregs *oscregs_pool = NULL;
static struct player *player_pool = NULL;

static struct osc *_osc_alloc(void)
{
  struct osc *osc;
  if (osc_pool) {
    osc = osc_pool;
    osc_pool = osc_pool->next;
  } else {
    osc = memory_alloc(sizeof(*osc));
  }
  return osc;
}

static void _osc_free(struct osc *osc)
{
  osc->next = osc_pool;
  osc_pool = osc;
}

static struct oscregs *_oscregs_alloc(void)
{
  struct oscregs *oscregs;
  if (oscregs_pool) {
    oscregs = oscregs_pool;
    oscregs_pool = oscregs_pool->next;
  } else {
    oscregs = memory_alloc(sizeof(*oscregs));
  }
  return oscregs;
}

static void _oscregs_free(struct oscregs *oscregs)
{
  oscregs->next = oscregs_pool;
  oscregs_pool = oscregs;
}

static struct player *_player_alloc(void)
{
  struct player *player;
  if (player_pool) {
    player = player_pool;
    player_pool = player_pool->next;
  } else {
    player = memory_alloc(sizeof(*player));
  }
  return player;
}

static void _player_free(struct player *player)
{
  player->next = player_pool;
  player_pool = player;
}

static int oscregs_free(struct oscregs *oscregs)
{
  struct oscregs *next;

  for (; oscregs; oscregs = next) {
    next = oscregs->next;
    _oscregs_free(oscregs);
  }

  return 0;
}

static struct oscregs *oscregs_alloc_num(int num)
{
  struct oscregs *oscregs = NULL, *regs;
  int i;

  for (i = 0; i < num; i++) {
    if ((regs = _oscregs_alloc()) == NULL)
      goto err;
    memset(regs, 0, sizeof(*regs));

    regs->next = oscregs;
    oscregs = regs;
  }

  return oscregs;

err:
  oscregs_free(oscregs);
  return NULL;
}

static int oscregs_push(struct oscregs **pool, struct oscregs *oscregs)
{
  struct oscregs *regs;

  while (oscregs) {
    regs = oscregs;
    oscregs = oscregs->next;

    regs->next = *pool;
    *pool = regs;
  }

  return 0;
}

static struct oscregs *oscregs_pop(struct oscregs **pool, struct oscset *oscset)
{
  struct oscregs *oscregs = NULL, *regs;
  struct osc *osc;

  for (osc = oscset->oscs; osc; osc = osc->next) {
    if ((regs = *pool) == NULL)
      goto err;
    *pool = (*pool)->next;
    memset(regs, 0, sizeof(*regs));

    regs->osc = osc;

    regs->next = oscregs;
    oscregs = regs;
  }

  return oscregs;

err:
  oscregs_push(pool, oscregs);
  return NULL;
}

static int step_clear(struct samplestep *step)
{
  step->sec   = 0;
  step->count = 0;
  return 0;
}

static int step_num(struct samplestep *step, int samplefreq)
{
  return step->sec * samplefreq + step->count;
}

static int step_diff(struct samplestep *step0, struct samplestep *step1, int samplefreq)
{
  return (step0->sec - step1->sec) * samplefreq + step0->count - step1->count;
}

static int step_copy(struct samplestep *dst, struct samplestep *src)
{
  dst->sec   = src->sec;
  dst->count = src->count;
  return 0;
}

static int step_add(struct samplestep *dst, struct samplestep *src, int count, int samplefreq)
{
  struct samplestep step;

  if (!src)
    src = dst;

  step_copy(&step, src);

  step.count += count;
  while (step.count >= samplefreq) {
    step.sec++;
    step.count -= samplefreq;
  }

  step_copy(dst, &step);

  return 0;
}

static struct player *player_alloc(int oscset_index, int amp, int freq,
				   struct samplestep *step, int stepbase, int steplen)
{
  struct player *player;

  if ((player = _player_alloc()) == NULL)
    return NULL;
  memset(player, 0, sizeof(*player));

  player->amp      = amp;
  player->freq     = freq;
  step_copy(&player->step, step);
  player->stepbase = stepbase;
  player->steplen  = steplen;

  player->oscset_index = oscset_index;

  return player;
}

static int player_free(struct player *player)
{
  struct player *next;

  for (; player; player = next) {
    if (player->nextplayer) {
      player_free(player->nextplayer);
      player->nextplayer = NULL;
    }
    next = player->next;
    oscregs_free(player->oscregs);
    oscregs_free(player->pool);
    _player_free(player);
  }

  return 0;
}

#define SCALENUM 12

static int scale2freq[SCALENUM] = {
  16744036, /* C9  */
  17739688, /* C#9 */
  18794545, /* D9  */
  19912127, /* D#9 */
  21096164, /* E9  */
  22350607, /* F9  */
  23679643, /* F#9 */
  25087708, /* G9  */
  26579501, /* G#9 */
  28160000, /* A9  */
  29834481, /* A#9 */
  31608531, /* B9  */
};

static int player_add(struct player **pp, int oscset_index,
		      int amp, int scale, int octave,
		      struct samplestep *step, int stepbase, int steplen)
{
  struct player *player;
  int freq = 0, i;

  if (scale >= 0) {
    freq = scale2freq[scale];
    for (i = octave; i < 9; i++)
      freq >>= 1;
    freq /= 1000;
  }

  player = player_alloc(oscset_index, amp, freq, step, stepbase, steplen);
  if (!player)
    return -1;

  player->next = *pp;
  *pp = player;

  return 0;
}

static int init(void)
{
  static int initialized = 0;

  if (!initialized) {
    memset(oscsets, 0, sizeof(oscsets));
#ifdef USE_SDL
    sdl_start();
#endif
    initialized = 1;
  }

  return 0;
}

int audio_check(void)
{
  int i;

  if (playeres)
    return NLL_ERRCODE_MEMORY_NOT_EMPTY;

  for (i = 0; i < AUDIO_OSCSET_MAXNUM; i++) {
    if (oscsets[i].oscs || oscsets[i].num)
      return NLL_ERRCODE_MEMORY_NOT_EMPTY;
  }

  return 0;
}

int audio_init(void)
{
  return 0;
}

int audio_done(void)
{
  int r;
  struct osc *osc;
  struct oscregs *oscregs;
  struct player *player;

  if ((r = player_free(playeres)) < 0)
    return r;
  playeres = NULL;

  if ((r = audio_clear_all()) < 0)
    return r;

  while (osc_pool) {
    osc = osc_pool;
    osc_pool = osc_pool->next;
    memory_free(osc);
  }

  while (oscregs_pool) {
    oscregs = oscregs_pool;
    oscregs_pool = oscregs_pool->next;
    memory_free(oscregs);
  }

  while (player_pool) {
    player = player_pool;
    player_pool = player_pool->next;
    memory_free(player);
  }

  return 0;
}

#if 0
#define AMPLIMIT 32767
#define SINNUM 64
static int sin_val[SINNUM + 1] = {
      0,   804,  1608,  2410,  3212,  4011,  4808,  5602,
   6393,  7179,  7962,  8739,  9512, 10278, 11039, 11793,
  12539, 13279, 14010, 14732, 15446, 16151, 16846, 17530,
  18204, 18868, 19519, 20159, 20787, 21403, 22005, 22594,
  23170, 23731, 24279, 24811, 25329, 25832, 26319, 26790,
  27245, 27683, 28105, 28510, 28898, 29268, 29621, 29956,
  30273, 30571, 30852, 31113, 31356, 31580, 31785, 31971,
  32137, 32285, 32412, 32521, 32609, 32678, 32728, 32757, 32767
};
#else
#define AMPLIMIT 32767
#define SINNUM 256
static int sin_val[SINNUM + 1] = {
      0,   201,   402,   603,   804,  1005,  1206,  1407,
   1608,  1809,  2009,  2210,  2410,  2611,  2811,  3012,
   3212,  3412,  3612,  3811,  4011,  4210,  4410,  4609,
   4808,  5007,  5205,  5404,  5602,  5800,  5998,  6195,
   6393,  6590,  6786,  6983,  7179,  7375,  7571,  7767,
   7962,  8157,  8351,  8545,  8739,  8933,  9126,  9319,
   9512,  9704,  9896, 10087, 10278, 10469, 10659, 10849,
  11039, 11228, 11417, 11605, 11793, 11980, 12167, 12353,
  12539, 12725, 12910, 13094, 13279, 13462, 13645, 13828,
  14010, 14191, 14372, 14553, 14732, 14912, 15090, 15269,
  15446, 15623, 15800, 15976, 16151, 16325, 16499, 16673,
  16846, 17018, 17189, 17360, 17530, 17700, 17869, 18037,
  18204, 18371, 18537, 18703, 18868, 19032, 19195, 19357,
  19519, 19680, 19841, 20000, 20159, 20317, 20475, 20631,
  20787, 20942, 21096, 21250, 21403, 21554, 21705, 21856,
  22005, 22154, 22301, 22448, 22594, 22739, 22884, 23027,
  23170, 23311, 23452, 23592, 23731, 23870, 24007, 24143,
  24279, 24413, 24547, 24680, 24811, 24942, 25072, 25201,
  25329, 25456, 25582, 25708, 25832, 25955, 26077, 26198,
  26319, 26438, 26556, 26674, 26790, 26905, 27019, 27133,
  27245, 27356, 27466, 27575, 27683, 27790, 27896, 28001,
  28105, 28208, 28310, 28411, 28510, 28609, 28706, 28803,
  28898, 28992, 29085, 29177, 29268, 29358, 29447, 29534,
  29621, 29706, 29791, 29874, 29956, 30037, 30117, 30195,
  30273, 30349, 30424, 30498, 30571, 30643, 30714, 30783,
  30852, 30919, 30985, 31050, 31113, 31176, 31237, 31297,
  31356, 31414, 31470, 31526, 31580, 31633, 31685, 31736,
  31785, 31833, 31880, 31926, 31971, 32014, 32057, 32098,
  32137, 32176, 32213, 32250, 32285, 32318, 32351, 32382,
  32412, 32441, 32469, 32495, 32521, 32545, 32567, 32589,
  32609, 32628, 32646, 32663, 32678, 32692, 32705, 32717,
  32728, 32737, 32745, 32752, 32757, 32761, 32765, 32766, 32767
};
#endif

static int degree_inrange(int deg)
{
  while (deg < 0)
    deg += (SINNUM*4);
  deg %= (SINNUM*4);
  return deg;
}

static int int_sin(int deg)
{
  deg = degree_inrange(deg);
  if (deg > (SINNUM*2))
    return -int_sin(deg - (SINNUM*2));
  if (deg > SINNUM)
    return int_sin((SINNUM*2) - deg);
  return sin_val[deg];
}

static int int_triangle(int deg)
{
  deg = degree_inrange(deg);
  if (deg > (SINNUM*2))
    return -int_triangle(deg - (SINNUM*2));
  if (deg > SINNUM)
    return int_triangle((SINNUM*2) - deg);
  return deg * AMPLIMIT / SINNUM;
}

static int int_pulse(int deg, int n)
{
  deg = degree_inrange(deg);
  return (deg < (SINNUM*4) * n / AUDIO_PER_NUM) ? AMPLIMIT : -AMPLIMIT;
}

static int int_sawup(int deg)
{
  deg = degree_inrange(deg);
  return (deg - (SINNUM*2)) * AMPLIMIT / (SINNUM*2);
}

static int int_sawdown(int deg)
{
  deg = degree_inrange(deg);
  return ((SINNUM*2) - deg) * AMPLIMIT / (SINNUM*2);
}

static unsigned int _rand0(unsigned int seed)
{
  seed ^= seed << 13;
  seed ^= seed >> 17;
  seed ^= seed <<  5;
  return seed;
}

static unsigned int _rand1(unsigned int seed)
{
  seed ^= seed <<  1;
  seed ^= seed << 10;
  seed ^= seed >>  3;
  return seed;
}

static unsigned int _rand(unsigned int seed)
{
  return _rand0(seed) ^ _rand1(seed);
}

static int int_random(int deg)
{
  return (int)(_rand(deg ? deg : (AMPLIMIT + 1)) % (AMPLIMIT * 2 + 1)) - AMPLIMIT;
}

static int int_noise(int deg)
{
  return (int)(_rand(deg / (SINNUM*4) + 1) % (AMPLIMIT * 2 + 1)) - AMPLIMIT;
}

static int int_zero(int deg)
{
  return 0;
}

static int int_high(int deg)
{
  return AMPLIMIT;
}

struct audio_data {
  volatile int finished;
  int num;
  int samplefreq;
  int channels;
  struct samplestep step;
  struct player *ready;
  struct player *busy;
  struct player *complete;
};

static int int_amp(int type, int deg)
{
  int a, param;

  param = type & ~NLL_A_TYPE_MASK;

  switch (type & NLL_A_TYPE_MASK) {
  case NLL_A_SIN:       a = int_sin(deg); break;
  case NLL_A_TRIANGLE:  a = int_triangle(deg); break;
  case NLL_A_SQUARE:    a = int_pulse(deg, AUDIO_PER_NUM / 2); break;
  case NLL_A_PULSE:     a = int_pulse(deg, param); break;
  case NLL_A_SAWUP:     a = int_sawup(deg); break;
  case NLL_A_SAWDOWN:   a = int_sawdown(deg); break;
  case NLL_A_RANDOM:    a = int_random(deg); break;
  case NLL_A_NOISE:     a = int_noise(deg); break;
  case NLL_A_ZERO:      a = int_zero(deg); break;
  case NLL_A_HIGH:      a = int_high(deg); break;
  default:              a = 0; break;
  }

  return a;
}

static struct osc *oscset_search(struct oscset *oscset, int index)
{
  struct osc *osc;

  for (osc = oscset->oscs; osc; osc = osc->next) {
    if (osc->index == index)
      break;
  }

  return osc;
}

static int oscreg_calc_envelope(struct oscregs *regs, int step)
{
  struct osc *osc;
  struct envpoint *p0;
  struct envpoint *p1;

  osc = regs->osc;

  if (!osc->envelope.num)
    return AUDIO_PER_NUM;

  p0 = &osc->envelope.points[regs->envelope.index];

  if (regs->envelope.index >= osc->envelope.num - 1)
    return p0->amp;

  p1 = p0 + 1;

  if (p0->step >= p1->step)
    return p0->amp;

  return p0->amp + (p1->amp - p0->amp) * (step - p0->step) / (p1->step - p0->step);
}

static int calc_deg(int step, int freq, int samplefreq)
{
  int deg, d, m;

#if 0
  deg = step * freq * (SINNUM*4) / samplefreq;
#else /* To avoid overflow, calculate division and modulo separately */
  d = (step * freq) / samplefreq;
  m = (step * freq) % samplefreq;
  deg  = d * (SINNUM*4);
  deg += m * (SINNUM*4) / samplefreq;
#endif

  return deg;
}

static int oscreg_calc(struct oscregs *oscregs, int amp, int freq,
		       int samplefreq, int step, int stepbase, int steplen)
{
  struct oscregs *regs;
  struct osc *osc;
  int i, deg, f, a, count = 0, outamp = 0;
  int envstep, envstep_fixed, envstep_linear;

  if (!freq)
    return 0;

  if (!oscregs) {
    deg = calc_deg(step, freq, samplefreq);
    count = 1;
    outamp = int_amp(NLL_A_SIN, deg);
  }

  for (regs = oscregs; regs; regs = regs->next) {
    osc = regs->osc;
    osc->in.num = 0;
    osc->in.amp = 0;
    osc->out.amp = 0;
  }

  for (regs = oscregs; regs; regs = regs->next) {
    osc = regs->osc;
    osc->out.amp = regs->out.amp;
    for (i = 0; i < osc->outnum; i++) {
      if (osc->outs[i].osc) {
	osc->outs[i].osc->in.num++;
	osc->outs[i].osc->in.amp += regs->out.amp * osc->outs[i].amp / AUDIO_PER_NUM;
      } else {
	count++;
	outamp += regs->out.amp * osc->outs[i].amp / AUDIO_PER_NUM;
      }
    }
  }

  for (regs = oscregs; regs; regs = regs->next) {
    osc = regs->osc;
    if (osc->in.num)
      osc->in.amp /= osc->in.num;
  }

  if (count)
    outamp /= count;

  envstep_linear = (steplen  > 0) ? (step * AUDIO_PER_NUM / steplen ) : 0;
  envstep_fixed  = (stepbase > 0) ? (step * AUDIO_PER_NUM / stepbase) : 0;

  for (regs = oscregs; regs; regs = regs->next) {
    osc = regs->osc;

    a = 0;

    if ((osc->type & NLL_A_TYPE_MASK) == NLL_A_EXTIN) {
      if (osc->in.osc)
	a = osc->in.osc->out.amp;
    } else {
      f = osc->freq;

      if ((osc->flags & NLL_A_MOD_MASK) == NLL_A_FREQ) {
	/*
	 * To avoid overflow, calculate division and modulo separately
	 * (f = f + osc->in.amp * f / AMPLIMIT)
	 */
	f = f +
	  (osc->in.amp * (f / 100) / (AMPLIMIT / 100)) +
	  (osc->in.amp * (f % 100) / AMPLIMIT);
      }

      if (osc->flags & NLL_A_FIXED) {
	deg = calc_deg(step, f, samplefreq);
      } else {
	deg = calc_deg(step, freq, samplefreq);
	/*
	 * To avoid overflow, calculate division and modulo separately
	 * (deg = deg * f / AUDIO_PER_NUM)
	 */
	deg = (deg * (f / 100) / (AUDIO_PER_NUM / 100)) + (deg * (f % 100) / AUDIO_PER_NUM);
      }

      deg += osc->phase * (SINNUM*4) / AUDIO_PER_NUM;

      if ((osc->flags & NLL_A_MOD_MASK) == NLL_A_PHASE) {
	deg = deg + osc->in.amp * (SINNUM*2) / AMPLIMIT;
      }

      a = int_amp(osc->type, deg);
    }

    a = a * osc->amp / AUDIO_PER_NUM;

    switch (osc->flags & NLL_A_MOD_MASK) {
    case NLL_A_NULL: a =  osc->in.amp; break;
    case NLL_A_REV:  a = -osc->in.amp; break;
    case NLL_A_SIGN: a = (osc->in.amp <  0) ? -a : ((osc->in.amp > 0) ? a :           0); break;
    case NLL_A_CLIP: a = (osc->in.amp < -a) ? -a : ((osc->in.amp > a) ? a : osc->in.amp); break;
    case NLL_A_ADD:  a =  osc->in.amp + a; break;
    case NLL_A_SUB:  a =  osc->in.amp - a; break;
    case NLL_A_MUL:  a =  osc->in.amp * a; break;
    case NLL_A_AND:  a =  osc->in.amp & a; break;
    case NLL_A_OR:   a =  osc->in.amp | a; break;
    case NLL_A_XOR:  a =  osc->in.amp ^ a; break;
    case NLL_A_AVE:  a = (osc->in.amp + a) / 2; break;
    case NLL_A_PER:  a = (osc->in.amp / 0x10) * (a / 0x10) / (AMPLIMIT / 0x100); break;
    case NLL_A_MIN:  a = (osc->in.amp < a) ? osc->in.amp : a; break;
    case NLL_A_MAX:  a = (osc->in.amp > a) ? osc->in.amp : a; break;
    case NLL_A_LT:   a = (osc->in.amp < a) ? AMPLIMIT : -AMPLIMIT; break;
    case NLL_A_GT:   a = (osc->in.amp > a) ? AMPLIMIT : -AMPLIMIT; break;
    case NLL_A_PHASE:
    case NLL_A_FREQ:
    default:
      break;
    }

    envstep = (osc->flags & NLL_A_FIXEDENV) ? envstep_fixed : envstep_linear;

    while ((regs->envelope.index + 1 < osc->envelope.num) &&
	   (osc->envelope.points[regs->envelope.index + 1].step <= envstep)) {
      regs->envelope.index++;
    }
    a = a * oscreg_calc_envelope(regs, envstep) / AUDIO_PER_NUM;

    regs->out.amp = a;
  }

  return outamp * amp / AUDIO_PER_NUM;
}

static int calc_frames(struct audio_data *audiodata, short *frames, int n)
{
  int i, amp;
  struct player *player, **pp;
  struct samplestep stepend;
  struct oscset *oscset;

  if (audiodata->finished) {
    for (i = 0; i < n; i++) {
      frames[i] = 0;
      step_add(&audiodata->step, NULL, 1, audiodata->samplefreq);
    }
    return n;
  }

  step_add(&stepend, &audiodata->step, n, audiodata->samplefreq);

  for (pp = &audiodata->ready; *pp;) {
    player = *pp;

    if (step_diff(&stepend, &player->step, audiodata->samplefreq) < 0) {
      pp = &player->nextplayer;
      continue;
    }

    if (player->next) {
      *pp = player->next;
      (*pp)->nextplayer = player->nextplayer;
    } else {
      *pp = player->nextplayer;
    }

    player->nextplayer = NULL;

    player->next = audiodata->busy;
    audiodata->busy = player;
  }

  for (i = 0; i < n; i++) {
    amp = 0;
    for (pp = &audiodata->busy; *pp;) {
      player = *pp;

      step_add(&stepend, &player->step, player->steplen, audiodata->samplefreq);

      if (!(step_diff(&audiodata->step, &stepend, audiodata->samplefreq) < 0)) {
	*pp = player->next;
	oscregs_push(&player->head->pool, player->oscregs);
	player->oscregs = NULL;
	player->next = audiodata->complete;
	audiodata->complete = player;
	continue;
      }

      if (!(step_diff(&audiodata->step, &player->step, audiodata->samplefreq) < 0)) {
	if (!player->oscregs) {
	  oscset = &oscsets[player->oscset_index];
	  player->oscregs = oscregs_pop(&player->head->pool, oscset);
	}
	amp += oscreg_calc(player->oscregs,
			   player->amp,
			   player->freq, audiodata->samplefreq,
			   step_diff(&audiodata->step, &player->step, audiodata->samplefreq),
			   player->stepbase, player->steplen);
      }
      pp = &player->next;
    }

    amp = audiodata->num ? (amp / audiodata->num) : 0;
    if (amp >  AMPLIMIT) amp =  AMPLIMIT;
    if (amp < -AMPLIMIT) amp = -AMPLIMIT;
    frames[i] = amp;
    step_add(&audiodata->step, NULL, 1, audiodata->samplefreq);
  }

  n = i;

  if (!audiodata->ready && !audiodata->busy)
    audiodata->finished = 1;

  return n;
}

#ifdef USE_SDL
static void sdl_callback(void *userdata, unsigned char *stream, int len)
{
  struct audio_data *audiodata;
  short *frames;
  int n;

  audiodata = (struct audio_data *)userdata;

  frames = (short *)stream;
  n = len / 2;

  calc_frames(audiodata, frames, n);
}
#endif

static int _run(struct audio_data *audiodata, struct samplestep *step)
{
#ifdef USE_SDL
  SDL_AudioSpec spec;
#endif

#ifdef USE_SDL
  memset(&spec, 0, sizeof(spec));

  spec.freq = audiodata->samplefreq;
  spec.format = AUDIO_S16;
  spec.channels = audiodata->channels;
  spec.samples = 256;
  spec.callback = sdl_callback;
  spec.userdata = audiodata;

  SDL_OpenAudio(&spec, NULL);
  SDL_PauseAudio(0);
#if 0
  SDL_Delay((step->sec * 1000) + (step->count * 1000 / audiodata->samplefreq));
#else
  while (!audiodata->finished) {
    SDL_Delay(10);
  }
#endif
  SDL_PauseAudio(1);
  SDL_CloseAudio();
#endif

  return 0;
}

static int _run_file(struct audio_data *audiodata, struct samplestep *step, const char *filename)
{
  FILE *fp;
  int r, bits = 16, offset = 0, n, samplenum;
#define FRAMENUM 256
  short frames[FRAMENUM];

  samplenum = step_num(step, audiodata->samplefreq);

  fp = sound_open(filename, audiodata->samplefreq, audiodata->channels, bits, samplenum);
  if (!fp)
    return -1;

  for (offset = 0; offset < samplenum;) {
    n = FRAMENUM;
    if (offset + n >= samplenum)
      n = samplenum - offset;
    calc_frames(audiodata, frames, n);
    sound_write(fp, audiodata->channels, bits, samplenum, &offset, frames, n);
    if (audiodata->finished)
      break;
  }

  if ((r = sound_close(fp, audiodata->channels, bits, samplenum, &offset)) < 0)
    return -1;

  return 0;
}

static int player_oscregs_maxnum(struct player *player)
{
  struct oscset *oscset;
  int num = 0;

  for (; player; player = player->next) {
    if (!player->freq)
      continue;
    oscset = &oscsets[player->oscset_index];
    if (num < oscset->num)
      num = oscset->num;
  }

  return num;
}

static struct player *player_run(struct player *player, struct samplestep *step, const char *filename)
{
  static struct audio_data audiodata;
  struct player *p;
  int num = 0;

  init();

  for (p = player; p; p = p->nextplayer) {
    num++;
    p->pool = oscregs_alloc_num(player_oscregs_maxnum(p) * (p->duplicate + 1));
  }

  audiodata.finished = 0;
  audiodata.num = num;
  audiodata.samplefreq = AUDIO_SAMPLING_FREQ;
  audiodata.channels = 1;
  step_clear(&audiodata.step);

  audiodata.ready    = player;
  audiodata.busy     = NULL;
  audiodata.complete = NULL;

  if (filename)
    _run_file(&audiodata, step, filename);
  else
    _run(&audiodata, step);

  audiodata.finished = 1;

  while (audiodata.ready) {
    player = audiodata.ready;
    audiodata.ready = audiodata.ready->nextplayer;

    player->nextplayer = audiodata.complete;
    audiodata.complete = player;
  }

  while (audiodata.busy) {
    player = audiodata.busy;
    audiodata.busy = audiodata.busy->nextplayer;

    player->nextplayer = audiodata.complete;
    audiodata.complete = player;
  }

  return audiodata.complete;
}

int audio_clear(int oscset_index)
{
  struct oscset *oscset;
  struct osc *next;

  init();

  if (oscset_index >= AUDIO_OSCSET_MAXNUM)
    return -1;

  if (oscset_index < 0)
    return audio_clear_all();

  oscset = &oscsets[oscset_index];

  player_free(playeres);
  playeres = NULL;

  for (; oscset->oscs; oscset->oscs = next) {
    next = oscset->oscs->next;
    _osc_free(oscset->oscs);
  }

  oscset->num = 0;

  return 0;
}

int audio_clear_all(void)
{
  int r, i;

  init();

  for (i = 0; i < AUDIO_OSCSET_MAXNUM; i++) {
    if ((r = audio_clear(i)) < 0)
      return r;
  }

  return 0;
}

int audio_osc(int output, int outamp, int amp, int freq, int phase, int type, int flags, int oscset_index)
{
  struct oscset *oscset;
  struct osc *osc;
  int r, n;

  init();

  if ((oscset_index < 0) || (oscset_index >= AUDIO_OSCSET_MAXNUM))
    return -1;

  oscset = &oscsets[oscset_index];

  if ((osc = _osc_alloc()) == NULL)
    return -1;
  memset(osc, 0, sizeof(*osc));

  osc->index = ++oscset->num;
  osc->type = type;
  osc->amp  = amp;
  osc->freq = freq;
  osc->phase = phase;
  osc->flags = flags;

  osc->next = oscset->oscs;
  oscset->oscs = osc;

  if (osc->flags & NLL_A_CLEARENV) {
    osc->envelope.num = 0;
  } else {
    n = 0;
    osc->envelope.points[n  ].step = 0;
    osc->envelope.points[n++].amp  = AUDIO_PER_NUM;
    osc->envelope.points[n  ].step = AUDIO_PER_NUM;
    osc->envelope.points[n++].amp  = 0;
    osc->envelope.num = n;
  }

  if ((osc->type & NLL_A_TYPE_MASK) == NLL_A_EXTIN) {
    osc->in.osc = oscset_search(oscset, osc->type & ~NLL_A_TYPE_MASK);
  }

  if ((r = audio_addout(osc->index, output, outamp, flags, oscset_index)) < 0) {
    oscset->oscs = osc->next;
    _osc_free(osc);
    return r;
  }

  return osc->index;
}

int audio_addout(int index, int output, int amp, int flags, int oscset_index)
{
  struct oscset *oscset;
  struct osc *osc, *out;

  init();

  if ((oscset_index < 0) || (oscset_index >= AUDIO_OSCSET_MAXNUM))
    return -1;

  oscset = &oscsets[oscset_index];

  osc = oscset_search(oscset, index);
  if (!osc)
    return -1;

  if (osc->outnum == AUDIO_OSC_OUTMAX)
    return -1;

  if (output < 0)
    return 0;

  if (output == NLL_A_OUTPUT) {
    out = NULL;
  } else {
    out = oscset_search(oscset, output);
    if (!out)
      return -1;
  }

  osc->outs[osc->outnum  ].amp = amp;
  osc->outs[osc->outnum++].osc = out;

  return 0;
}

int audio_envpoint(int index, int n, int step, int amp, int flags, int oscset_index)
{
  struct oscset *oscset;
  struct osc *osc;

  init();

  if ((oscset_index < 0) || (oscset_index >= AUDIO_OSCSET_MAXNUM))
    return -1;

  oscset = &oscsets[oscset_index];

  osc = oscset_search(oscset, index);
  if (!osc)
    return -1;

  if (flags & NLL_A_CLEARENV) {
    osc->envelope.num = 0;
  }

  if (n < 0)
    n = osc->envelope.num;

  if (n >= AUDIO_OSC_ENVMAX)
    return -1;

  osc->envelope.points[n].step = step;
  osc->envelope.points[n].amp  = amp;
  n++;

  if (osc->envelope.num < n)
    osc->envelope.num = n;

  return 0;
}

int audio_envcopy(int index, int source, int flags, int oscset_index)
{
  struct oscset *oscset;
  struct osc *osc, *sosc;
  int i;

  init();

  if ((oscset_index < 0) || (oscset_index >= AUDIO_OSCSET_MAXNUM))
    return -1;

  oscset = &oscsets[oscset_index];

  osc = oscset_search(oscset, index);
  if (!osc)
    return -1;

  sosc = oscset_search(oscset, source);
  if (!sosc)
    return -1;

  for (i = 0; i < AUDIO_OSC_ENVMAX; i++) {
    if (i < sosc->envelope.num) {
      osc->envelope.points[i].step = sosc->envelope.points[i].step;
      osc->envelope.points[i].amp  = sosc->envelope.points[i].amp;
    } else {
      osc->envelope.points[i].step = 0;
      osc->envelope.points[i].amp  = 0;
    }
  }

  osc->envelope.num = sosc->envelope.num;

  return 0;
}

static int code2scale[] = { 0, 2, 4, 5, 7, 9, 11 };

int audio_play(const char *string, int flags, int oscset_index, const char *filename)
{
  static struct player **pp;
  int scale, octave = 3, oct, len;
  int interval, steplen;
  static int length = 4, amp = AUDIO_PER_NUM, tempo = 120, stepmod = 0;
  static struct samplestep step, max;
  int bracket = 0, bracketlen = 0, duplicate = 0;
  const char *s, *upper, *lower;
  char *p;

  init();

  if ((oscset_index < 0) || (oscset_index >= AUDIO_OSCSET_MAXNUM))
    return -1;

  upper = "CDEFGAB";
  lower = "cdefgab";

  if (!playeres) {
    step_clear(&max);
  }

  if (!playeres || (flags & NLL_A_NEW)) {
    pp = &playeres;
    step_clear(&step);
    stepmod = 0;
    duplicate = 0;
  }

  for (s = string; *s;) {
    switch (*s) {
    case ' ':
    case '\t':
    case '\r':
    case '\n':
      s++;
      break;

    case 'O':
      s++;
      if (*s && isdigit(*s)) {
	octave = *s - '0';
	s++;
      }
      break;

    case '<':
      s++;
      octave--;
      break;

    case '>':
      s++;
      octave++;
      break;

    case 'L':
      s++;
      if (*s && isdigit(*s)) {
	length = strtol(s, &p, 10);
	if (s == p)
	  goto err;
	s = p;
	length = length ? length : 4;
      }
      break;

    case 'V':
      s++;
      if (*s && isdigit(*s)) {
	amp = strtol(s, &p, 10);
	if (s == p)
	  goto err;
	s = p;
	amp = amp ? amp : AUDIO_PER_NUM;
      }
      break;

    case 'T':
      s++;
      if (*s && isdigit(*s)) {
	tempo = strtol(s, &p, 10);
	if (s == p)
	  goto err;
	s = p;
	tempo = tempo ? tempo : 120;
      }
      break;

    case '(':
      s++;
      if (bracket)
	goto err;
      bracketlen = 0;
      bracket = 1;
      break;

    case ')':
      s++;
      if (!bracket)
	goto err;
      step_add(&step, NULL, bracketlen, AUDIO_SAMPLING_FREQ);
      bracket = 0;
      duplicate = 0;
      break;

    case 'R':
    case 'r':
    default:
      if ((*s == 'R') || (*s == 'r')) {
	scale = -1;
      } else if ((p = strchr(upper, *s)) != NULL) {
	scale = code2scale[p - upper];
      } else if ((p = strchr(lower, *s)) != NULL) {
	scale = code2scale[p - lower];
      } else {
	goto err;
      }
      s++;

      if (scale < 0) {
	oct = -1;
      } else {
	oct = octave;

	if (*s && ((*s == '#') || (*s == '+'))) {
	  scale++;
	  s++;
	}

	if (*s && (*s == '-')) {
	  scale--;
	  s++;
	}

	while (scale < 0) {
	  scale += SCALENUM;
	  oct--;
	}

	while (scale >= SCALENUM) {
	  scale -= SCALENUM;
	  oct++;
	}
      }

      len = length;
      if (*s && isdigit(*s)) {
	len = strtol(s, &p, 10);
	if (s == p)
	  goto err;
	s = p;
      }

      if (len <= 0)
	goto err;
      interval = 64 / len;

      if (*s && (*s == '.')) {
	interval = interval * 3 / 2;
	s++;
      }

#if 0
      msec = (1000 * 60 / 16) * interval / tempo;
      steplen = msec * AUDIO_SAMPLING_FREQ / 1000;
#else /* To avoid the accumulation of the error, save the remainder */
      steplen = 60 * interval * AUDIO_SAMPLING_FREQ / 16;
      steplen += stepmod; /* add the old remainder */
      stepmod = steplen % tempo; /* save the remainder */
      steplen = steplen / tempo;
#endif

      if (player_add(pp, oscset_index, amp, scale, oct,
		     &step, AUDIO_SAMPLING_FREQ * 60 / tempo, steplen) < 0)
	goto err;
      if (pp == &playeres) {
	(*pp)->nextplayer = (*pp)->next;
	(*pp)->next = NULL;
      }
      (*pp)->head = playeres;
      pp = &(*pp)->next;

      if (!bracket) {
	step_add(&step, NULL, steplen, AUDIO_SAMPLING_FREQ);
      } else {
	if (bracketlen < steplen)
	  bracketlen = steplen;
	if (playeres->duplicate < duplicate)
	  playeres->duplicate = duplicate;
	duplicate++;
      }

      break;
    }
  }

  if (step_diff(&max, &step, AUDIO_SAMPLING_FREQ) < 0)
    step_copy(&max, &step);

  if (!(flags & NLL_A_SAVE)) {
    if (playeres) {
      playeres = player_run(playeres, &max, filename);
      if (!playeres)
	return -1;
      player_free(playeres);
      playeres = NULL;
    }
  }

  return 0;

err:
  return -1;
}

#endif
